/*
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>	/* Header */
#include "rt_tool.h"
#include "ralink.h"

const char *strAryNetworkType[]=
{
    "Ad Hoc",
    "Infrastructure", 
    "Automatic"
};

void site_survey(char *buf)
{
  char strSSID[128];
  PNDIS_802_11_BSSID_LIST_EX plssid;
  PNDIS_WLAN_BSSID_EX     pBssid;
  int i, nSigQua;

  plssid = (PNDIS_802_11_BSSID_LIST_EX) buf;
  printf("sites=%d\n", (int)plssid->NumberOfItems);
  pBssid = (PNDIS_WLAN_BSSID_EX) plssid->Bssid;
  for (i = 0; i < (int)plssid->NumberOfItems; i++) {
        memset(strSSID, 0x00, NDIS_802_11_LENGTH_SSID + 1);
        memcpy(strSSID, pBssid->Ssid.Ssid, pBssid->Ssid.SsidLength);
	printf("ssid=%s,", strSSID);
	printf("channel=%d,", pBssid->Channel);
        printf("type=%s, ", strAryNetworkType[pBssid->InfrastructureMode]);
	printf("privacy=%d ", (int)pBssid->Privacy);	
        nSigQua = ConvertRssiToSignalQuality(pBssid->Rssi);
        printf("signal quality=(%d%%)(%d) ", nSigQua, (int)pBssid->Rssi);
	printf("\n");
        pBssid = (PNDIS_WLAN_BSSID_EX)((char *)pBssid + pBssid->Length);
  }
}

int get_strongest_AP(char *buf)
{
  char strSSID[128];
  PNDIS_802_11_BSSID_LIST_EX plssid;
  PNDIS_WLAN_BSSID_EX     pBssid;
  int i, nSigQua = 0, tmp;

  plssid = (PNDIS_802_11_BSSID_LIST_EX) buf;
  if((int)plssid->NumberOfItems == 0) {
	return -1;
  } 
  pBssid = (PNDIS_WLAN_BSSID_EX) plssid->Bssid;
  for (i = 0; i < (int)plssid->NumberOfItems; i++) {
        tmp = ConvertRssiToSignalQuality(pBssid->Rssi);
	if(tmp > nSigQua) {
		nSigQua = tmp;
        	memset(strSSID, 0x00, NDIS_802_11_LENGTH_SSID + 1);
        	memcpy(strSSID, pBssid->Ssid.Ssid, pBssid->Ssid.SsidLength);
	}
        pBssid = (PNDIS_WLAN_BSSID_EX)((char *)pBssid + pBssid->Length);
	
  }
  printf("%s", strSSID); 
  return 0;
}

int
main(int argc, char **	argv)
{
  int skfd;		/* generic raw socket desc.	*/
  int goterr = 0;
  struct iwreq		wrq;
  char strSSID[128];
  int cs;

  if(argc == 1) {	// list scan
	printf("iwpriv list: list all APs\n");
        printf("iwpriv set ssid auth_mode enc_type [key_id] key\n");
	return 0;
  }

  /* Create a channel to the NET kernel. */
  if((skfd = iw_sockets_open()) < 0)
    {
      perror("socket");
      return(-1);
    }
  strcpy(wrq.ifr_name, "ra0");

  if(!strcmp(argv[1], "list")) {
  	char buf[1024];

	if((goterr = RL_LIST_SCAN(skfd, &wrq, buf)) < 0)
		goto close_fd;
	
	if(argc == 2)
		site_survey(buf);
	else if(argc == 3 && !strcmp(argv[2], "strongest")) {
		goterr = get_strongest_AP(buf);
	}
	goto close_fd;
  } 
  else if(!strcmp(argv[1], "status")) {
	int cs;

  	RL_Connect_Status(skfd, &wrq, &cs);
	goto close_fd;
  } 
  strcpy(strSSID, argv[2]);

  // disable then enable
  RL_Enable_WLAN(skfd, &wrq, 0);
  RL_Enable_WLAN(skfd, &wrq, 1);
  // remove key
  RL_Remove_Key(skfd, &wrq);
  // set essid
  RL_Set_SSID(skfd, &wrq, strSSID);
  if(argc == 3) {	// non-encryption
	RL_Set_Auth_Mode(skfd, &wrq, 0);
	RL_Set_Encryption_Type(skfd, &wrq, 0);
	goto close_fd;
  }
  // set auth mode
  // set encryption type
  // set key id(for WEP)
  // sey key(WEP or PSK)
  if(!strcmp(argv[3], "open")) {
	int keyid=1;
	char *key;

	RL_Set_Auth_Mode(skfd, &wrq, 0);	// open
	if(argc == 4)
		goto close_fd;
	if(strcmp(argv[4], "wep")) {
		printf("Need Encryption type: wep.\n");
		goto close_fd;
	}
	RL_Set_Encryption_Type(skfd, &wrq, 1);
	key = argv[5];
	if(argc > 6) {
		keyid = atoi(argv[5]);
		key = argv[6];
	}
	RL_Set_Default_Key_ID(skfd, &wrq, keyid);
	RL_Set_WEP_Key(skfd, &wrq, key);
	
  }
  else if(!strcmp(argv[3], "wpapsk")) {
  	int type = 2;

	RL_Set_Auth_Mode(skfd, &wrq, 2);	// wpapsk
	if(argc < 4) {
		printf("Need Encryption type: tkip, aes.\n");
		goto close_fd;
	}
	if(strcmp(argv[4], "tkip") && strcmp(argv[4], "aes")) {
		printf("Need Encryption type: tkip, aes.\n");
		goto close_fd;
	}
	if(!strcmp(argv[4], "tkip"))
		type = 2;
	if(!strcmp(argv[4], "aes"))
		type = 3;
	RL_Set_Encryption_Type(skfd, &wrq, type);
	if(argc < 5) {
		printf("Need Encryption key: 8~63 chars\n");
		goto close_fd;
	}
	RL_Set_PSK_Key(skfd, &wrq, argv[5]);
  }
  // check connected ?
  RL_Connect_Status(skfd, &wrq, &cs);
  // get rssi
  RL_Get_RSSI(skfd, &wrq);

close_fd:
  /* Close the socket. */
  close(skfd);

  return(goterr);
}
